/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucla.edu, imarkov@cs.ucla.edu
***  Affiliation:   UCLA, Computer Science Department,
***                 Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/




//Base Small Placer
// - Small Placement Problem
//   created by Andrew Caldwell  981206

#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

#include "sRowBBPlacer.h"

BranchSmallPlacer::BranchSmallPlacer(SmallPlacementProblem& problem,
					Parameters params)
	: BaseSmallPlacer(problem, params), 
	  _c(problem.getNumCells()+2),
	  _curSoln(problem.getNumCells())
{
    abkfatal(problem.getNumRows() == 1, 
	"BB SmallPlacer can't deal with multiple rows");


    Placement& soln         = _problem.getSolnBuffer()[0].placement;
    //vector<Orient>& ori     = _problem.getSolnBuffer()[0].orientations;
    const int numCells = _problem.getNumCells();

	//setup the stack and queue to produce the ordering
    _s.reserve(numCells);

    _c[numCells]   = numCells;
    _c[numCells+1] = numCells+1;
    _curWL = _netStacks->getCost(); //this will be just the WL associated
				   //with terminals + any minWL/net

    for(_idx = numCells; _idx != 0; )
   	 _q.push_back(--_idx); 

    bool usingInitSoln = false;
    double bestWL = DBL_MAX;

    unsigned n;
    for(n = numCells; n > 0; )
	soln[--n].y = problem.getRows()[0].yMin;

    if(_problem.getSolnBuffer()[0].populated && _params.useInitSoln)
    {
	std::sort(_q.begin(), _q.end(), CompareNodesByInitPl(soln));
	usingInitSoln = true;
    }

	//generate the initial bound
    while(!_q.empty())
    {
        _s.push_back(_q.front());
        _q.pop_front();
        const unsigned c = _s.back();
        _curSoln[c].x = _netStacks->getLeftEdge();
        _curWL = _netStacks->addNodeToLeft(c);
    }

    bestWL = _curWL;
    unsigned c;
    for(c = numCells; c != 0; )
    {   c--; soln[c].x = _curSoln[c].x;}
    _problem.getSolnBuffer()[0].populated = true;

    while(!_s.empty())
    {
        c = _s.back();
        _q.push_front(c);  //restors the orig. order
        _s.pop_back();
        _curWL = _netStacks->removeNodeFromLeft(c);
    }

    if(_params.verb.forMajStats > 5)
	   cout<<" Orig Bound: "<<bestWL<<endl;

	//sort(_q.begin(), _q.end(), CompareNodesByDegree(problem));

    _idx = numCells-1;

    unsigned numAdds = 0;

    while(_idx < numCells)
    {
    	addCell();
	numAdds++;

        if(_c[_idx] == 0 || _curWL >= bestWL)
      	{
	    _c[_idx] = 0;	//force a bound

	    if(_curWL < bestWL) //got here if:
				// new best complete soln (curWL < best)
				// bounded partial soln (curWL > best)
				//  so there is no need to additionally
				//  check to ensure this is a complete soln
	    { 
		if(_params.verb.forMajStats > 5)
		   cout<<" New Best: "<<_curWL<<" found after "<<numAdds<<endl;

		bestWL = _curWL;

                unsigned cell;
		for(cell = numCells; cell != 0; )
		{ cell--; soln[cell].x = _curSoln[cell].x;}
    	        _problem.getSolnBuffer()[0].populated = true;

	    }

            while(_c[_idx] == 0)	//|| _netStacks->getCost()==_BestWL
            {
	   	if(_idx < numCells)
		  removeCell();

    		_c[++_idx]--; 
            }
      	}
      	_idx--;
    }

    if(_params.verb.forMajStats > 5)
    {
	cout<<" Total Add Operations: "<<numAdds<<endl;
        cout<<"Final soln has cost: "<<bestWL<<endl;
    }
}

void BranchSmallPlacer::addCell()
{
    _s.push_back(_q.front());
    _q.pop_front();
    _c[_idx] = _idx;

    const unsigned cell = _s.back();

    if(_idx%2)
    {
      _curWL = _netStacks->addNodeToRight(cell);
      _curSoln[cell].x = _netStacks->getRightEdge();
    }
    else
    {
      _curSoln[cell].x = _netStacks->getLeftEdge();
      _curWL = _netStacks->addNodeToLeft(cell);
    }
}

void BranchSmallPlacer::addCellLeft()
{
    _s.push_back(_q.front());
    _q.pop_front();
    _c[_idx] = _idx;

    const unsigned cell = _s.back();

    _curSoln[cell].x = _netStacks->getLeftEdge();
    _curWL = _netStacks->addNodeToLeft(cell);
}

void BranchSmallPlacer::removeCell()
{
    unsigned cell = _s.back();

    _q.push_back(cell);
    _s.pop_back();

    if(_idx%2)
      _curWL = _netStacks->removeNodeFromRight(cell);
    else
      _curWL = _netStacks->removeNodeFromLeft(cell);
}

void BranchSmallPlacer::removeCellLeft()
{
    unsigned cell = _s.back();

    _q.push_back(cell);
    _s.pop_back();

    _curWL = _netStacks->removeNodeFromLeft(cell);
}

void BranchSmallPlacer::print()
{
    cout<<"Q: ";
    unsigned i;
    for(i = 0; i < _q.size(); i++)
	cout<<_q[i]<<" ";
    cout<<endl; 

    cout<<"S: ";
    for(i = 0; i < _s.size(); i++)
	cout<<_s[i]<<" ";
    cout<<endl; 

    cout<<"C: ";
    for(i = _c.size(); i != 0; )
	cout<<_c[--i]<<" ";
    cout<<endl;

    cout<<"Idx:   "<<_idx<<endl;
    cout<<"CurWL: "<<_curWL<<endl<<endl;

    _netStacks->print();

    cout<<"__________________________"<<endl<<endl;
}


