/**************************************************************************
***
***    Copyright (c) 1995-1997 by ABKGroup, UCLA VLSI CAD Laboratory,
***    UCLA Computer Science Department, Los Angeles, CA 90095-1596 USA,
***    and by the Regents of the University of California.
***    All rights reserved.
***
***    No portion of this file may be used, copied, or transmitted for
***    any purpose, including any research or commercial purpose,
***    without the express written consent of Professor Andrew B. Kahng,
***    UCLA Computer Science Department, 3713 Boelter Hall, Los Angeles,
***    CA 90095-1596 USA.   Any copy made of this file must include this
***    notice.    ABKGroup and the Regents of the University of California
***    retain title to all copies of this file and/or its compiled
***    derivatives.
***
***************************************************************************/


//created by Andrew Caldwell on 10/17/99 caldwell@cs.ucla.edu


#include "splitSmall.h"
#include "partProbForCapo.h"
#include "mlPart.h"
#include "fmPart.h"
#include "bbFMPart.h"

SplitSmallBlockVertically::SplitSmallBlockVertically
	(      CapoBlock&  blockToSplit,
	 const CapoPlacer& capo,
	       CapoSplitterParams params)
	: BaseBlockSplitter(blockToSplit,capo,params)
{
    if(_params.verb.forActions > 2)
	cout<<"Splitting a small block Vertically"<<endl;
   
    //1)construct the partitioning problem
    //2)while(!foundAGoodSplit)
    //    decide which partitioner to use, and partition the block
    //    if needed, change tolerances and repartition
    //	  see if this is a good split (legal, etc)
    //3)construct new blocks


    double   capacity  = _block.getCapacity();
    double   targetCap = capacity/2.0;
    vector<double> partCaps;

    _xSplit = _block.findXSplit(targetCap,targetCap,0,partCaps);
    PartitioningProblemForCapo problem(_block, capo, false, _xSplit);

    callPartitioner(problem);

    if(_params.verb.forActions > 8)
	cout<<"Done partitioning...creating blocks"<<endl;

    createVSplitBlocks(problem, _xSplit);
}
 
    
void SplitSmallBlockVertically::callPartitioner
		(PartitioningProblemForCapo& problem)
{
    bool foundGoodSplit = false;
    unsigned numTries   = 0;
   
    while(!foundGoodSplit && numTries++ < 2)
    {
	if(_block.getNumCells() <= _capo.getParams().smallPartThreshold)
	{
	    if(_params.verb.forActions > 8)
		cout<<"Using BBPart"<<endl;
	    foundGoodSplit = callBBFM(problem);
	}
	else
	{
	    if(_params.verb.forActions > 8)
		cout<<"Using FMPart"<<endl;
	    foundGoodSplit = callFMPart(problem);

	    if(foundGoodSplit)
	    {
	 	if(_params.verb.forActions > 9)
		    cout<<"  Success.  Found a good partitioning"<<endl;
	    }
	    else if(numTries < 2)
	    {
	 	if(_params.verb.forActions > 9)
		    cout<<"  Failed. clearing part buffers"<<endl;
		PartitioningBuffer& buff = problem.getSolnBuffers();
		unsigned numTerms=problem.getHGraph().getNumTerminals();
		unsigned sol;
	        for(sol = buff.beginUsedSoln(); 
		    sol != buff.endUsedSoln(); sol++)
		{
		    Partitioning& part = buff[sol];
	 	    for(unsigned i = numTerms; i < part.size(); i++)
			part[i].setToAll(2);
		}

		//consider changing the problem here to attempt to
		//get a better solution	
	    }
	    else if(_params.verb.forActions > 9)
		cout<<"  Failed too many times. Taking the best soln"<<endl;
	}
    }
}

bool SplitSmallBlockVertically::callBBFM
		(PartitioningProblemForCapo& problem)
{
    problem.setTolerance(10);

    BBFMPart::Parameters bbfmparams;
    bbfmparams.fmStarts = 5;
    bbfmparams.bbParams.pushLimit = 1000000; //~2 seconds

    problem.reserveBuffers(1);
    problem.getSolnBuffers().setBeginUsedSoln(0);
    problem.getSolnBuffers().setEndUsedSoln(1);

    BBFMPart partitioner1(problem, bbfmparams);

    vector<double> partAreas(2,0.0);
    vector<double> siteAreas(2,0.0);
    const Partitioning& soln = problem.getBestSoln();
    for(unsigned c = 0; c < problem.getHGraph().getNumNodes(); c++)
    {
	unsigned part    = soln[c].isInPart(0) ? 0 : 1;
	partAreas[part] += problem.getHGraph().getNodeByIdx(c).getWeight();
    }

    _xSplit = _block.findXSplit(partAreas[0], partAreas[1], 10, siteAreas);

	//really ought to be more pro-active here.......
    if(  _block.getRelativeWhitespace() > 0 &&
      ( (siteAreas[0] < partAreas[0] || siteAreas[1] < partAreas[1])))
    {
         _xSplit = _block.findXSplit(partAreas[0], partAreas[1], 0, siteAreas);
    }

    return true; 
}

bool SplitSmallBlockVertically::callFMPart
		(PartitioningProblemForCapo& problem)
{
    problem.setTolerance(20.0);

    FMPartitioner::Parameters fmParams;
    fmParams.useEarlyStop = false;
    fmParams.useClip = (_block.getNumCells() > 50) ? 1 : 0;
    fmParams.verb.forActions  /= 4;
    fmParams.verb.forMajStats /= 4;
    fmParams.verb.forSysRes   /= 4;

    unsigned numFMStarts = 4;
    problem.reserveBuffers(numFMStarts);
    problem.getSolnBuffers().setBeginUsedSoln(0);
    problem.getSolnBuffers().setEndUsedSoln(numFMStarts);
   

    FMPartitioner partitioner1(problem, fmParams);

    vector<double> partAreas(2,0.0);
    vector<double> siteAreas(2,0.0);
    const Partitioning&  soln = problem.getBestSoln();
    const HGraphFixed& hgraph = problem.getHGraph();

    for(unsigned c = hgraph.getNumTerminals(); c < hgraph.getNumNodes(); c++)
    {
	unsigned part    = soln[c].isInPart(0) ? 0 : 1;
	partAreas[part] += hgraph.getNodeByIdx(c).getWeight();
    }

		//for small blocks this isn't really the best
		//way to do this.  Really, we ought to compute the
		//range of splits that give us some min ws% in
		//each new block, and then pick the best xSplit
		//within that range.
    double idealXSplit = _block.findXSplit(partAreas[0], partAreas[1], 
						0, siteAreas);
    _xSplit            = _block.findXSplit(partAreas[0], partAreas[1], 
						5, siteAreas);

    if(_params.verb.forMajStats > 9)
    {
	cout<<" CellAreas: "<<partAreas[0]<<"/"<<partAreas[1]<<endl;
	cout<<" Total:     "<<_block.getTotalCellArea()<<endl;
	cout<<" SiteAreas: "<<siteAreas[0]<<"/"<<siteAreas[1]<<endl;
    }

		//cutline was shifted
    if(_xSplit != idealXSplit && _params.doRepartitioning)	
    {
      double origCut = 
	problem.getCosts()[problem.getBestSolnNum()];

      problem.setTolerance(2);

      if(_params.verb.forMajStats > 4)
      {
        cout<<" Repartitioning FMPart"<<endl;
        cout<<"Targets:  "<<problem.getCapacities()[0][0]<<"/"
                          <<problem.getCapacities()[1][0]<<endl;
        cout<<"Max:      "<<problem.getMaxCapacities()[0][0]<<"/"
                          <<problem.getMaxCapacities()[1][0]<<endl;
        cout<<"Min:      "<<problem.getMinCapacities()[0][0]<<"/"
                          <<problem.getMinCapacities()[1][0]<<endl;
      }        

      fmParams.useClip = false;
      FMPartitioner partitioner2(problem, fmParams);

      double newCut = 
	problem.getCosts()[problem.getBestSolnNum()];

      double cutDelta = newCut - origCut;
      if(cutDelta > (origCut/10) && cutDelta >= 5)
      {
	if(_params.verb.forActions > 9)
	  cout<<" ->FM Failed.  Repartitioning increased "
	      <<"the cut too much"<<endl;
 	return false;
      }	
    }
    else
	_xSplit = idealXSplit;

    unsigned bestSolnNum = problem.getBestSolnNum(); 
    if(_params.verb.forMajStats > 4)
	cout<<" Best Solution had cut "<<problem.getCosts()[bestSolnNum]<<endl;
  
    if(problem.getViolations()[bestSolnNum])
    {
	if(_params.verb.forActions > 9)
	  cout<<" ->FM Failed.  No legal solutions"<<endl;
	_xSplit = idealXSplit;
 	return false;
    }

    return true;
}

